part of '../flutterx_live_data.dart';

/// Builds a [Widget] when given a concrete value of a [LiveData].
typedef LiveWidgetBuilder<T> = Widget Function(BuildContext context, T value);

/// A [LiveData] observer that is also a [Widget] whose content is built from the LiveData value
///
/// Every time [data] is invalidated the widget is rebuilt with the latest value
///
/// [T] The type of data held by [data]
class LiveDataBuilder<T> extends StatelessWidget {
  /// LiveData which is observed by this builder
  final LiveData<T> data;

  /// [LiveWidgetBuilder] of this widget content
  final LiveWidgetBuilder<T> builder;

  /// Create a [LiveDataBuilder] of [data] from given [builder]
  const LiveDataBuilder({Key? key, required this.data, required this.builder}) : super(key: key);

  @override
  StatelessElement createElement() => _LiveDataBuilderElement<T>(this);

  @override
  Widget build(BuildContext context) => data.isActive ? builder(context, data._data) : const SizedBox.shrink();
}

class _LiveDataBuilderElement<T> extends StatelessElement {
  _LiveDataBuilderElement(LiveDataBuilder widget) : super(widget);

  @override
  LiveDataBuilder<T> get widget => super.widget as LiveDataBuilder<T>;

  @override
  void mount(Element? parent, dynamic newSlot) {
    super.mount(parent, newSlot);
    widget.data.observeForever(_onChanged);
  }

  @override
  void unmount() {
    widget.data.removeObserver(_onChanged);
    super.unmount();
  }

  void _onChanged(T value) => markNeedsBuild();
}

/// Equals to [LiveDataBuilder] but allows to have the [builder] type different from source [data] type by
/// applying a [transform] to a [MediatorLiveData]
///
/// [S] The type of data held by [data]
/// [T] The transformed value type
class MediatorLiveDataBuilder<S, T> extends StatelessWidget {
  /// LiveData which is observed by this builder
  final LiveData<S> data;

  /// Transform applied to [data]
  final T Function(S) transform;

  /// [LiveWidgetBuilder] of this widget content
  final LiveWidgetBuilder<T> builder;

  /// Create a [MediatorLiveDataBuilder] of [transform]ed [data] from given [builder]
  const MediatorLiveDataBuilder({Key? key, required this.data, required this.transform, required this.builder})
      : super(key: key);

  @override
  StatelessElement createElement() => _MediatorLiveDataBuilderElement<S, T>(this);

  @override
  Widget build(BuildContext context) {
    final data = (context as _MediatorLiveDataBuilderElement<S, T>)._mediator;
    return data.isActive ? builder(context, data._data) : const SizedBox.shrink();
  }
}

class _MediatorLiveDataBuilderElement<S, T> extends StatelessElement {
  final MediatorLiveData<T> _mediator = MediatorLiveData<T>();

  _MediatorLiveDataBuilderElement(MediatorLiveDataBuilder widget) : super(widget);

  @override
  MediatorLiveDataBuilder<S, T> get widget => super.widget as MediatorLiveDataBuilder<S, T>;

  @override
  void mount(Element? parent, dynamic newSlot) {
    super.mount(parent, newSlot);
    _mediator.addTransformSource(widget.data, widget.transform);
    _mediator.observeForever(_onChanged);
  }

  @override
  void unmount() {
    _mediator.removeObserver(_onChanged);
    _mediator.removeSource(widget.data);
    super.unmount();
  }

  void _onChanged(T value) => markNeedsBuild();
}
